/* factor.c -- factorization of very smooth numbers.
 *
 * Copyright (C) 2012, 2013 INRIA
 *
 * This file is part of CMH.
 *
 * CMH is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 *
 * CMH is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>
#include "factor.h"

static factor_matrix factor_matrix_alloc (mpz_srcptr n)
    /* (over-)allocates a matrix f to hold the factorisation of n,
       where f [i, 0] is the i-th prime factor and f [i, 1] its
       multiplicity; the last entry is indicated by a prime factor 0 */
{
    int omega = mpz_sizeinbase (n, 2) - 1;
    factor_matrix f;

    f = (factor_matrix) malloc ((omega + 1) * 2 * sizeof (unsigned long));

    return f;
}


factor_matrix factor_matrix_dup(factor_matrix f)
{
    int size;
    for (size = 0; f [size][0] ; size++);
    factor_matrix g = (factor_matrix) malloc ((size + 1) * 2 * sizeof (unsigned long));
    memcpy(g, f, (size + 1) * 2 * sizeof (unsigned long));
    return g;
}



void factor_matrix_free (factor_matrix f)
{
    free (f);
}


static int remove_prime_power (mpz_ptr n, unsigned long int p)
    /* return the p-valuation of n, 
       and remove the p-power part from n                            */
{
    int e = 0;
    for( ; mpz_divisible_ui_p (n, p) ; mpz_divexact_ui (n, n, p)) {
        e++;
    }

   return e;
}


static int coprime (unsigned long int a, unsigned long int b)
{
   if (b == 0)
      return 0;
   else if (b == 1)
      return 1;
   else
      return coprime (b, a%b);
}


factor_matrix factor_trialdiv (mpz_srcptr n, unsigned long int bound)
   /* factors n by trial division with numbers <= bound, or indefinitely
         if bound is 0.
      The return value is a newly allocated factor list if the number is
      smooth, NULL otherwise.  */
{
   const unsigned long int parray [] = { 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31 };
   const int length = 10; /* not counting the last prime, which is the
      starting point for the general loop */
   const unsigned long int lcm = 1078282205lu; /* without 2 and 3, handled
      separately, and without the last prime */
   int d = 2 - ((parray [length] - parray [length - 1]) % 6) / 2;
   mpz_t nloc;
   
   int c = 0;
   unsigned long int p;
   int i, ok;

   mpz_init_set (nloc, n);

   if (bound == 0)
      bound = ULONG_MAX;

   factor_matrix f = factor_matrix_alloc (nloc);
   
   for (i = 0; i < length; i++) {
      f[c][0] = parray [i];
      f[c][1] = remove_prime_power (nloc, f[c][0]);
      if (f[c][1]) c++;
   }

   p = parray [length];
   while (p <= bound && mpz_cmp_ui (nloc, 1) > 0) {
      if (coprime (p, lcm)) {
          f[c][0] = p;
          f[c][1] = remove_prime_power (nloc, p);
          if (f[c][1]) c++;
      }
      p += 2 + ((d & 1) << 1);
         /* avoid numbers divisible by 2 or 3 */
      d++;
   }
  
   f [c][0] = 0;
   f = realloc (f, (c+1) * 2 * sizeof (unsigned long));

   ok = (mpz_cmp_ui (nloc, 1) == 0);
   mpz_clear (nloc);

   if (ok)
       return f;
   else {
       factor_matrix_free(f);
       return NULL;
   }
}


void factor_matrix_asprintf (char ** pres, factor_matrix f)
{
   size_t alloc = 10;
   size_t size = 0;
   char * res = malloc (alloc);
   int d, i;
   
#define PUSH(fmt__, arg__) do {                                         \
    for( ; (d = snprintf(res + size, alloc-size, fmt__, arg__)) >= (int) (alloc-size) ; ) {     \
        alloc = size+d+1 + alloc / 4;                                   \
        res = realloc(res, alloc);                                      \
    }                                                                   \
    size+=d;                                                            \
} while (0)

   if (f[0][0] == 0) {
       snprintf(res, alloc, "1");
       *pres = res;
       return;
   }
   for (i = 0; f [i][0] != 0; i++) {
      if (i > 0)
         PUSH ("%c", '*');
      PUSH ("%lu", f [i][0]);
      if (f [i][1] > 1)
         PUSH ("^%lu", f [i][1]);
   }
#undef PUSH

   *pres = res;
}


int factor_matrix_fprintf (FILE *file, factor_matrix f)
{
   int ok, i, size;

   for (size = 0; f [size][0] != 0; size++);
   ok = fprintf (file, "%i", size);
   for (i = 0; i < size; i++)
      ok += fprintf (file, " %lu %lu", f [i][0], f [i][1]);

   return (ok > 0);
}


void mpz_remove_squares(mpz_ptr D)
{
    if (mpz_cmp_ui(D, 1000000000) > 0) {
        fprintf(stderr, "Warning, square removal is completely dumb !\n");
    }

    mpz_t R;
    mpz_init_set_ui(R, 1);
    int u, s;
    u = 3;
        mpz_t S;
        mpz_init(S);
        mpz_sqrt(S, D);
        s = mpz_get_ui(S);
    for ( ; mpz_fdiv_ui(D, 4) == 0 ; mpz_fdiv_q_2exp(D, D, 2), s>>=1 );
    for( ; u <= s ; u+=2) {
        for( ; u <= s && ! mpz_divisible_ui_p(D, u) ; u+=2);
        if (u > s)
            break;
        int m = 0;
        for( ; mpz_divisible_ui_p(D, u) ; m++) mpz_divexact_ui(D, D, u);
        if (m&1) mpz_mul_ui(R, R, u);
        mpz_sqrt(S, D);
        s = mpz_get_ui(S);
    }
    mpz_mul(D, D, R);
    mpz_clear(S);
    mpz_clear(R);
}

void factor_matrix_fold_inner(mpz_ptr z, factor_matrix f, int i0, int i1)
{
    if (i1 - i0 == 1) {
        mpz_ui_pow_ui(z, f[i0][0], f[i0][1]);
    } else {
        int j = (i0 + i1) / 2;
        mpz_t l,r;
        mpz_init(l);
        mpz_init(r);
        factor_matrix_fold_inner(l, f, i0, j);
        factor_matrix_fold_inner(r, f, j, i1);
        mpz_mul(z, l, r);
        mpz_clear(l);
        mpz_clear(r);
    }
}

void factor_matrix_fold(mpz_ptr z, factor_matrix f)
{
   int size;
   for (size = 0; f [size][0] != 0; size++);
    if (size == 0) {
        mpz_set_ui(z, 1);
    } else {
        factor_matrix_fold_inner(z, f, 0, size);
    }
}

