/* quadratic.c -- functions for algebraic numbers in quadratic number fields
 *
 * Copyright (C) 2010, 2011, 2012 INRIA
 *
 * This file is part of CMH.
 *
 * CMH is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 *
 * CMH is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#include "quadratic_field.h"
#include "recognize_quadratic.h"

void quadratic_number_init(quadratic_number_ptr q)
{
    mpz_init(q->a);
    mpz_init(q->b);
    mpz_init(q->c);
}
void quadratic_number_clear(quadratic_number_ptr q)
{
    mpz_clear(q->a);
    mpz_clear(q->b);
    mpz_clear(q->c);
}
void quadratic_number_swap(quadratic_number_ptr q, quadratic_number_ptr r)
{
    mpz_swap(q->a, r->a);
    mpz_swap(q->b, r->b);
    mpz_swap(q->c, r->c);
}
void quadratic_number_set(quadratic_number_ptr q, quadratic_number_srcptr r)
{
    mpz_set(q->a, r->a);
    mpz_set(q->b, r->b);
    mpz_set(q->c, r->c);
}

static void
quadratic_number_normalize(quadratic_number_ptr q, quadratic_number_srcptr r)
{
    quadratic_number_set(q, r);
    mpz_t g;
    mpz_init(g);
    mpz_gcd(g, q->a, q->c);
    mpz_gcd(g, q->b, g);
    if (mpz_cmp_ui(g, 1) > 0) {
        mpz_divexact(q->a, q->a, g);
        mpz_divexact(q->b, q->b, g);
        mpz_divexact(q->c, q->c, g);
    }
    mpz_clear(g);
    if (mpz_cmp_ui(q->c, 0) < 0) {
        mpz_neg(q->c, q->c);
        mpz_neg(q->a, q->a);
        mpz_neg(q->b, q->b);
    }
}

static void
quadratic_number_mul_z(quadratic_number_ptr q, quadratic_number_srcptr r, mpz_srcptr z)
{
    mpz_mul(q->a, r->a, z);
    mpz_mul(q->b, r->b, z);
    mpz_set(q->c, r->c);
}

void
quadratic_number_mul_z_normalize(quadratic_number_ptr q, quadratic_number_srcptr r, mpz_srcptr z)
{
    mpz_t g,h;
    mpz_init(g);
    mpz_init(h);
    mpz_gcd(g, r->c, z);
    quadratic_number_set(q, r);
    mpz_divexact(q->c, q->c, g);
    mpz_divexact(h, z, g);
    quadratic_number_mul_z(q, q, h);
    quadratic_number_normalize(q, q);
    mpz_clear(g);
    mpz_clear(h);
}

/* Since it's useful only for _real_ quadratics, we don't bother providing an
 * interface for imaginary quadratic fields at all */
int quadratic_number_recognize(quadratic_number_ptr q, mpfr_srcptr x, mpz_srcptr disc)
{
    int res = recognize_quadratic(q->a, q->b, q->c, x, disc);
    quadratic_number_normalize(q, q);
    return res;
}
