/* recognize_quadratic.c -- reconstruct elements of quadratic number fields
 *
 * Copyright (C) 2010, 2011, 2012 INRIA
 *
 * This file is part of CMH.
 *
 * CMH is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 *
 * CMH is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#include <mpfr.h>
#include <assert.h>
#include <float.h>
#include <fplll.h>
#include <fplll/matrix.h>
#include "recognize_quadratic.h"
#include "cm2.h"
#include "io.h"

using namespace fplll;

/* try to recognize an algebraic number in a quadratic field */
/* The quadratic field is assumed to be that of fundamental discriminant
 * disc.
 *
 * This code attempts to find integers a,b,c, such that x is close to
 * (a+b*sqrt(disc))/c
 * 
 * returns true on success.
 */
int recognize_quadratic(mpz_ptr a, mpz_ptr b, mpz_ptr c, mpfr_srcptr x, mpz_srcptr disc)
{
    assert(!mpfr_nan_p(x));
    mpfr_prec_t prec = mpfr_get_prec(x);
    if (mpfr_get_exp(x) < - 0.9*prec) {
        /* What if we have, as an input, 1.1237651e-598, for precision
         * 2000 ? Without this guard, we're unable to recognize 0...
         */
        mpz_set_si(a, 0);
        mpz_set_si(b, 0);
        mpz_set_si(c, 1);
        return 1;
    }
    mpfr_t s; // sqrt(disc)
    mpfr_init2(s, prec);
    mpfr_set_z(s, disc, GMP_RNDN);
    mpfr_sqrt(s, s, GMP_RNDN);
    
    ZZ_mat<mpz_t> M(3,3);

    mpfr_t ur;
    mpfr_init2(ur, prec);
    for(int i = 0 ; i < 3 ; i++) {
        for(int j = 0 ; j < 3 ; j++) {
            M.Get(i,j).set(0);
        }
    }

    mp_exp_t X = mpfr_get_exp(x);
    mp_exp_t Xp = X > 0 ? X : 0;
    mp_exp_t Xm = X > 0 ? 0 : -X;
    assert(X == Xp-Xm);
    mp_exp_t E = prec - X;

    mpz_t u[3];
    for (int i = 0; i < 3; i++)
       mpz_init (u [i]);
    mpz_t scale_b, scale_c;
    mpz_init (scale_b);
    mpz_init (scale_c);

    /*
     * Note that in (a+b*s)/c \approx x, the magnitudes of a, b, and c
     * are unbalanced. Typically, we expect a/c \approx s*b/c \approx x
     * (where in fact the magnitude of s is negligible). Following the crude
     * approximation idea, we have x \approx 2^(Xp-Xm) (only one of Xp
     * and Xm is nonzero), thus:
     *  a*2^Xm \approx b*s*2^Xm \approx c*2^Xp.
     *
     * An adequately ``small'' solution (a,b,c) should then have
     * (a*2^Xm,b*2^Xm,c*2^Xp) roughly of the same magnitude.
     *
     * The error |x-(a+b*s)/c| should be a small multiple of
     * ulp(x) \approx 2^(-E), and |a+b*s-c*x| \approx |c|*2^Xp*2^(-E-Xp).
     * Scaling the last column by 2^(E+Xp), all entries in the target
     * vector should have about the same order of magnitude.
     *
     * XXX Note that this construction is badly sensible to garbage bits
     * in the tail of the floating point expansion of x. If x is correct
     * up to precision lambda, but is effectively an mpfr_t with
     * announced precision lambda + 1000, then the ``small'' vector won't
     * be as small as we expect.
     *
     * Here is the matrix fed to LLL:
     *
     * 2^Xm     0      0     2^(E+Xp)
     *  0    s*2^Xm    0   s*2^(E+Xp)
     *  0       0     2^Xp x*2^(E+Xp)
     *
     * In the case x has positive exponent Xp, this leads to the following
     * matrix:
     *
     *  1   0   0     2^(E+Xp)
     *  0   s   0   s*2^(E+Xp)
     *  0   0  2^Xp x*2^(E+Xp)
     *
     * And for negative exponent -Xm:
     *
     * 2^Xm    0    0    2^E
     *  0   s*2^Xm  0  s*2^E
     *  0       0   1  x*2^E
     * 
     * We may then drop the first column, which results in a faster LLL and
     * some more effort to derive a.
     */

    mpz_ui_pow_ui(u[0], 2, E+Xp);
    M.Get(0,2).set(u[0]);

    mpfr_mul_2ui(ur, s, Xm, MPFR_RNDN);
    mpfr_get_z(scale_b, ur, MPFR_RNDN);
    M.Get(1,0).set(scale_b);

    mpfr_mul_2ui(ur, s, E+Xp, MPFR_RNDN);
    mpfr_get_z(u[1], ur, MPFR_RNDN);
    M.Get(1,2).set(u[1]);

    mpz_ui_pow_ui(scale_c, 2, Xp);
    M.Get(2,1).set(scale_c);

    mpfr_mul_2ui (ur, x, E+Xp, GMP_RNDN);
    mpfr_get_z(u[2], ur, MPFR_RNDN);
    M.Get(2,2).set(u[2]);

    int res = (lllReduction (M) == 0);
    Z_NR<mpz_t> * vect = M.GetVec(0);

    if (mpz_cmp_ui(vect[1].GetData(),0) == 0)
        res = 0;

    if (res) {
        mpz_t tmp1, tmp2;
        mpz_init (tmp1);
        mpz_init (tmp2);

        mpz_set(b, vect[0].GetData());
        mpz_neg(c, vect[1].GetData());

        assert(mpz_divisible_p(b, scale_b));
        mpz_divexact(b, b, scale_b);
        assert(mpz_divisible_p(c, scale_c));
        mpz_divexact(c, c, scale_c);

        mpz_mul (tmp1, c, u [2]);
        mpz_mul (tmp2, b, u [1]);
        mpz_sub (tmp1, tmp1, tmp2);
        mpz_add (tmp1, tmp1, vect [2].GetData ());
        assert (mpz_divisible_p (tmp1, u [0]));
        mpz_divexact (a, tmp1, u [0]);

        mpz_clear (tmp1);
        mpz_clear (tmp2);
    }

    mpfr_clear(ur);
    mpfr_clear(s);
    for (int i = 0; i < 3; i++)
       mpz_clear (u [i]);
    mpz_clear (scale_b);
    mpz_clear (scale_c);
    return res;
}
